#
# Copyright (c) 2017-2024 AutoDeployAI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pypmml.base import JavaModelWrapper


class Application(JavaModelWrapper):
    @property
    def name(self):
        return self.call('name')

    @property
    def version(self):
        return self.call('version')


class Header(JavaModelWrapper):

    @property
    def copyright(self):
        return self.call('copyright')

    @property
    def description(self):
        return self.call('description')

    @property
    def modelVersion(self):
        return self.call('modelVersion')

    @property
    def application(self):
        app = self.call('application')
        return Application(app) if app is not None else None




#
# Copyright (c) 2017-2024 AutoDeployAI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pypmml.base import JavaModelWrapper


class Field(JavaModelWrapper):

    def __init__(self, java_model):
        super(Field, self).__init__(java_model)

    @property
    def name(self):
        return self.call('name')

    @property
    def displayName(self):
        return self.call('displayName')

    @property
    def dataType(self):
        return self.call('dataType').toString()

    @property
    def opType(self):
        return self.call('opType').toString()

    @property
    def valuesAsString(self):
        return self.call('valuesAsString')


class OutputField(Field):

    def __init__(self, java_model):
        super(Field, self).__init__(java_model)

    @property
    def feature(self):
        return self.call('feature')

    @property
    def targetField(self):
        return self.call('targetField')

    @property
    def value(self):
        return self.call('value')

    @property
    def ruleFeature(self):
        return self.call('ruleFeature').toString()

    @property
    def algorithm(self):
        return self.call('algorithm').toString()

    @property
    def rank(self):
        return self.call('rank')


class DataDictionary(JavaModelWrapper):
    def __init__(self, java_model):
        super(DataDictionary, self).__init__(java_model)

    @property
    def fields(self):
        return [Field(x) for x in self.call('fields')]

    @property
    def fieldNames(self):
        return self.call('fieldNames')

    def get(self, name):
        fld = self.get(name)
        return Field(fld) if fld is not None else None



#
# Copyright (c) 2024 AutoDeployAI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC, abstractmethod
import os

class PMMLError(Exception):
    """Base exception of PyPMML"""

class JVMGateway(ABC):
    """Base class of JVM gateway"""
    def __init__(self):
        # Java path
        self.java_path = None

        # JVM options
        # Fix IllegalAccessError: cannot access class jdk.internal.math.FloatingDecimal
        self.java_opts = ["--add-exports=java.base/jdk.internal.math=ALL-UNNAMED"]
        java_opts = os.environ.get("JAVA_OPTS")
        if java_opts:
            self.java_opts.extend(java_opts.split())

        # Classpath
        jars_dir = os.environ["PYPMML_JARS_DIR"] if "PYPMML_JARS_DIR" in os.environ else \
            os.path.join(os.path.dirname(os.path.abspath(__file__)), 'jars')
        self.classpath = os.path.join(jars_dir, "*")

    @abstractmethod
    def launch_gateway(self, java_opts=None, java_path=None):
        """Launch a `Gateway` in a new Java process.
        :param java_opts: an array of extra options to pass to Java (the classpath
            should be specified using the `classpath` parameter, not `javaopts`.)
        :param java_path: If None, JVM gateway will use $JAVA_HOME/bin/java if $JAVA_HOME
            is defined, otherwise it will use "java".
        """
        if java_opts:
            self.java_opts.extend(java_opts)
        self.java_path = java_path

    def call_java_func(self, func, *args):
        result = func(*args)
        return self.java2py(result)

    @abstractmethod
    def java2py(self, r):
        return r

    @abstractmethod
    def call_java_static_func(self, class_name, func_name, *args):
        return None

    @abstractmethod
    def detach(self, java_object):
        pass

    @abstractmethod
    def shutdown(self):
        pass

    @abstractmethod
    def name(self):
        return "undefined"


class JPypeGateway(JVMGateway):
    """JPype: """
    import jpype
    import jpype.imports
    import importlib
    from jpype import JArray, JObject

    JavaException = None

    def __init__(self):
        super().__init__()
        JPypeGateway._gateway = None

    def launch_gateway(self, java_opts=None, java_path=None):
        """Launch a `Gateway` in a new Java process.
        :param java_opts: an array of extra options to pass to Java (the classpath
            should be specified using the `classpath` parameter, not `javaopts`.)
        :param java_path: If None, JVM gateway will use $JAVA_HOME/bin/java if $JAVA_HOME
            is defined, otherwise it will use "java".
        """
        super().launch_gateway(java_opts=java_opts, java_path=java_path)
        self.jpype.startJVM(*self.java_opts, jvmpath=self.java_path, classpath=self.classpath, convertStrings=True)

    def java2py(self, r):
        if isinstance(r, self.JArray):
            return [self.java2py(x) for x in r]
        elif isinstance(r, self.JObject):
            cls_name = r.getClass().getName()
            if cls_name == 'scala.Some':
                return r.get()
            elif cls_name == 'scala.None$':
                return None
            elif cls_name == 'scala.Enumeration$Val':
                return r.toString()
        return r

    def py2java(self, arg):
        if isinstance(arg, list):
            from java.util import ArrayList
            array_list = ArrayList()
            for x in arg:
                array_list.add(self.py2java(x))
            return array_list
        else:
            return arg

    def call_java_func(self, func, *args):
        java_args = [self.py2java(x) for x in args]
        result = func(*java_args)
        return self.java2py(result)

    def call_java_static_func(self, class_name, func_name, *args):
        try:
            class_obj = self.importlib.import_module(class_name)
            func_obj = getattr(class_obj, func_name)
            return func_obj(*args)
        except self.jpype.JException as e:
            raise PMMLError(e.message())

    def detach(self, java_object):
        pass

    def shutdown(self):
        self.jpype.shutdownJVM()

    def name(self):
        return "JPype"

class Py4jGateway(JVMGateway):
    """Py4j"""
    from py4j.java_collections import JavaArray
    from py4j.java_gateway import JavaObject
    from py4j.protocol import Py4JJavaError

    _gateway = None
    _jvm = None

    def __init__(self):
        super().__init__()
        Py4jGateway._gateway = None

    def launch_gateway(self, java_opts=None, java_path=None):
        """Launch a `Gateway` in a new Java process.
        :param java_opts: an array of extra options to pass to Java (the classpath
            should be specified using the `classpath` parameter, not `java_opts`.)
        :param java_path: If None, Py4J will use $JAVA_HOME/bin/java if $JAVA_HOME
            is defined, otherwise it will use "java".
        """
        from py4j.java_gateway import JavaGateway, launch_gateway, GatewayParameters

        super().launch_gateway(java_opts=java_opts, java_path=java_path)
        if Py4jGateway._gateway is None:
            _port = launch_gateway(classpath=self.classpath, javaopts=self.java_opts, java_path=self.java_path, die_on_exit=True)
            Py4jGateway._gateway = JavaGateway(gateway_parameters=GatewayParameters(port=_port, auto_convert=True))
            Py4jGateway._jvm = Py4jGateway._gateway.jvm

    def shutdown(self):
        if Py4jGateway._gateway is not None:
            Py4jGateway._gateway.shutdown()
            Py4jGateway._gateway = None

    def detach(self, java_object):
        Py4jGateway._gateway.detach(java_object)

    def java2py(self, r):
        if isinstance(r, self.JavaArray):
            return [self.java2py(x) for x in r]
        elif isinstance(r, self.JavaObject):
            cls_name = r.getClass().getName()
            if cls_name == 'scala.Some':
                return r.get()
            elif cls_name == 'scala.None$':
                return None
            elif cls_name == 'scala.Enumeration$Val':
                return r.toString()
        return r

    def call_java_static_func(self, class_name, func_name, *args):
        """ Call Java Static Function """
        try:
            class_obj = getattr(self._jvm, class_name)
            func_obj = getattr(class_obj, func_name)
            return func_obj(*args)
        except self.Py4JJavaError as e:
            je = e.java_exception
            raise PMMLError(je.getClass().getSimpleName(), je.getMessage())

    def name(self):
        return "Py4j"


#
# Copyright (c) 2017-2024 AutoDeployAI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
PyPMML is the Python API for PMML4S.
"""

from pypmml.model import Model
from pypmml.jvm import PMMLError
from pypmml.base import PMMLContext
from pypmml.version import __version__

__all__ = ['Model', 'PMMLError', 'PMMLContext']



from setuptools import setup
import os

VERSION_PATH = os.path.join("pypmml", "version.py")
exec(open(VERSION_PATH).read())

VERSION = __version__ # noqa

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="pypmml",
    version=VERSION,
    description="Python PMML scoring library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=["pypmml", "pypmml.jars"],
    package_data={
        "pypmml.jars": ["*.jar"]
    },
    # include_package_data=True,
    install_requires=[
        "py4j>=0.10.7", "JPype1"
    ],
    url="https://github.com/autodeployai/pypmml",
    download_url = "https://github.com/autodeployai/pypmml/archive/v" + VERSION + ".tar.gz",
    author="AutoDeployAI",
    author_email="autodeploy.ai@gmail.com",
    license="Apache License 2.0",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



#
# Copyright (c) 2017-2024 AutoDeployAI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from threading import RLock
from .jvm import JVMGateway

class PMMLContext(object):
    _gateway: JVMGateway = None
    _active_pmml_context = None
    _lock = RLock()

    def __init__(self, gateway_instance=None, gateway="py4j", java_opts=None, java_path=None):
        PMMLContext._ensure_initialized(
            self,
            gateway_instance=gateway_instance,
            gateway=gateway,
            java_opts=java_opts,
            java_path=java_path)

    @classmethod
    def _ensure_initialized(cls, instance, gateway_instance=None, gateway="py4j", java_opts=None, java_path=None):
        """
        Checks whether a Gateway of JVM is initialized or not.
        """
        with PMMLContext._lock:
            if not PMMLContext._gateway:
                PMMLContext._gateway = gateway_instance or cls.launch_gateway(
                    gateway=gateway, java_opts=java_opts, java_path=java_path
                )

            if instance:
                if PMMLContext._active_pmml_context and PMMLContext._active_pmml_context != instance:
                    # Raise error if there is already a running PMML context
                    raise ValueError("Cannot run multiple PMMLContexts at once")
                else:
                    PMMLContext._active_pmml_context = instance

    @classmethod
    def getOrCreate(cls, gateway="py4j", java_opts=None, java_path=None) -> 'PMMLContext':
        """
        Get or instantiate a PMMLContext and register it as a singleton object.
        :param java_opts: an array of extra options to pass to Java (the classpath
            should be specified using the `classpath` parameter, not `java_opts`.)
        :param java_path: If None, JVM will use $JAVA_HOME/bin/java if $JAVA_HOME
            is defined, otherwise it will use "java".
        :param gateway: JVM gateway engine, support one of ["py4j", "jpype"]
        """
        with PMMLContext._lock:
            if PMMLContext._active_pmml_context is None:
                PMMLContext(gateway=gateway, java_opts=java_opts, java_path=java_path)
            return PMMLContext._active_pmml_context

    @classmethod
    def launch_gateway(cls, gateway="py4j", java_opts=None, java_path=None) -> 'JVMGateway':
        """Launch a `Gateway` in a new Java process.
        :param gateway: JVM gateway engine, support one of ["py4j", "jpype"]
        :param java_opts: an array of extra options to pass to Java (the classpath
            should be specified using the `classpath` parameter, not `java_opts`.)
        :param java_path: If None, JVM will use $JAVA_HOME/bin/java if $JAVA_HOME
            is defined, otherwise it will use "java".
        :return: An object of `Gateway`
        """
        if isinstance(gateway, str) and gateway.lower() == "jpype":
            from .jvm import JPypeGateway
            jvm_gateway = JPypeGateway()
        else:
            from .jvm import Py4jGateway
            jvm_gateway = Py4jGateway()
        jvm_gateway.launch_gateway(java_opts=java_opts, java_path=java_path)
        return jvm_gateway

    @classmethod
    def shutdown(cls):
        """Shuts down the :class:`GatewayClient` and the
           :class:`CallbackServer <py4j.java_callback.CallbackServer>`.
        """
        with PMMLContext._lock:
            if PMMLContext._gateway:
                PMMLContext._gateway.shutdown()
            PMMLContext._gateway = None
            PMMLContext._active_pmml_context = None

    def call_java_static_func(self, class_name, func_name, *args):
        return self._gateway.call_java_static_func(class_name, func_name, *args)

    def call_java_func(self, func, *args):
        return self._gateway.call_java_func(func, *args)

    def detach(self, java_model):
        if self._gateway:
            self._gateway.detach(java_model)

    @classmethod
    def gateway(cls):
        return cls._gateway.name() if cls._gateway is not None else None

class JavaModelWrapper(object):
    """
    Wrapper for the model in JVM
    """
    def __init__(self, java_model):
        self._pc = PMMLContext.getOrCreate()
        self._java_model = java_model

    def __del__(self):
        if self._pc:
            self._pc.detach(self._java_model)

    def call(self, name, *args):
        return self._pc.call_java_func(getattr(self._java_model, name), *args)

    def __str__(self):
        return self.call('toString')




#
# Copyright (c) 2017-2024 AutoDeployAI
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os

from pypmml.base import JavaModelWrapper, PMMLContext
from pypmml.jvm import PMMLError
from pypmml.elements import Header
from pypmml.metadata import Field, OutputField, DataDictionary


class Model(JavaModelWrapper):
    """A PMML model.
    """
    def __init__(self, java_model):
        super(Model, self).__init__(java_model)

    @property
    def version(self):
        """PMML version."""
        return self.call('version')

    @property
    def header(self):
        """The header of this model."""
        return Header(self.call('header'))

    @property
    def dataDictionary(self):
        """The data dictionary of this model."""
        return DataDictionary(self.call('dataDictionary'))

    @property
    def modelElement(self):
        """Model element type."""
        return self.call('modelElement').toString()

    @property
    def functionName(self):
        """
        Describe the kind of mining model, e.g., whether it is intended to be used for clustering or for classification.
        """
        return self.call('functionName')

    @property
    def modelName(self):
        """
        Identifies the model with a unique name in the context of the PMML file.
        This attribute is not required. Consumers of PMML models are free to manage the names of the models at their
        discretion.
        """
        return self.call('modelName')

    @property
    def algorithmName(self):
        """
        The algorithm name is free-type and can be any description for the specific algorithm that produced the model.
        This attribute is for information only.
        """
        return self.call('algorithmName')

    @property
    def inputNames(self):
        """All input names."""
        return self.call('inputNames')

    @property
    def inputFields(self):
        """All input fields."""
        return [Field(x) for x in self.call('inputFields')]

    @property
    def targetName(self):
        """The target name."""
        return self.call('targetName')

    @property
    def targetNames(self):
        """All target names."""
        return self.call('targetNames')

    @property
    def targetField(self):
        """The target field."""
        return [Field(x) for x in self.call('targetField')]

    @property
    def targetFields(self):
        """All target fields."""
        return [Field(x) for x in self.call('targetFields')]

    @property
    def outputNames(self):
        """All output names."""
        return self.call('outputNames')

    @property
    def outputFields(self):
        """All output fields."""
        return [OutputField(x) for x in self.call('outputFields')]

    @property
    def classes(self):
        """The class labels in a classification model."""
        return self.call('classes')

    def setSupplementOutput(self, value):
        self.call('setSupplementOutput', value)
        return self

    def _is_nd_array(self, data):
        try:
            import numpy as np
            return isinstance(data, np.ndarray)
        except ImportError:
            return False

    def _is_pandas_dataframe(self, data):
        try:
            import pandas as pd
            return isinstance(data, pd.DataFrame)
        except ImportError:
            return False

    def _is_pandas_series(self, data):
        try:
            import pandas as pd
            return isinstance(data, pd.Series)
        except ImportError:
            return False

    def predict(self, data):
        """
        Predict values for a given data.

        :param data:
          Support dict, string in JSON, and Series, DataFrame of Pandas
        :return:
          Scoring results in the same format as input data
        """
        if isinstance(data, (dict, str, u"".__class__)):
            return self.call('predict', data)
        else:
            if isinstance(data, list):
                if data:
                    record = data[0]
                    if isinstance(record, list):
                        return [self.call('predict', record) for record in data]
                    else:
                        return self.call('predict', data)
                else:
                    return []
            elif self._is_nd_array(data):
                import numpy as np
                if data.ndim == 1:
                    return self.call('predict', data.tolist())
                elif data.ndim == 2:
                    try:
                        import pandas as pd
                        data = pd.DataFrame(data, columns=self.inputNames)
                        json_data = data.to_json(orient='split', index=False)
                        result = self.call('predict', json_data)
                        return pd.read_json(result, orient='split').values
                    except:
                        return [self.call('predict', record.tolist()) for record in data]
                else:
                    raise PMMLError('Max 2 dimensions are supported')
            elif self._is_pandas_dataframe(data):
                import pandas as pd
                json_data = data.to_json(orient='split', index=False)
                result = self.call('predict', json_data)
                return pd.read_json(result, orient='split')
            elif self._is_pandas_series(data):
                import pandas as pd
                record = data.to_dict()
                result = self.call('predict', record)
                return pd.DataFrame.from_records([result]).iloc[0]
            else:
                raise PMMLError('Data type "{type}" not supported'.format(type=type(data).__name__))

    @classmethod
    def fromFile(cls, name):
        """Load a model from PMML file with given pathname"""
        pc = PMMLContext.getOrCreate()
        java_model = pc.call_java_static_func("org.pmml4s.model.Model", "fromFile", name)
        return cls(java_model)

    @classmethod
    def fromString(cls, s):
        """Load a model from PMML in a string"""
        pc = PMMLContext.getOrCreate()
        java_model = pc.call_java_static_func("org.pmml4s.model.Model", "fromString", s)
        return cls(java_model)

    @classmethod
    def fromBytes(cls, bytes_array):
        """Load a model from PMML in an array of bytes"""
        pc = PMMLContext.getOrCreate()
        java_model = pc.call_java_static_func("org.pmml4s.model.Model", "fromBytes", bytes_array)
        return cls(java_model)

    @classmethod
    def load(cls, f):
        """Load a model from PMML in any formats of readable, a file path, a string,
        or an array of bytes(bytes or bytearray)"""
        model_content = f
        if hasattr(f, 'read') and callable(f.read):
            model_content = f.read()

        if isinstance(model_content, (bytes, bytearray)):
            model_content = model_content.decode('utf-8')

        if isinstance(model_content, (str, u"".__class__)):
            # Check if a file path
            if os.path.exists(model_content):
                model = cls.fromFile(model_content)
            else:
                model = cls.fromString(model_content)
            return model
        else:
            raise PMMLError('Input type "{type}" not supported'.format(type=type(f).__name__))

    @classmethod
    def close(cls):
        """Shutdown the gateway of JVM"""
        PMMLContext.shutdown()
